/*
 * eth_ecat_datagram.hpp
 *
 *  Created on: May 2, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_

#include <cstdint>

#include "ethernet/eth.hpp"
#include "ethernet_industry/ethercattype.hpp"
#include "ethernet_industry/eth_ecat_types.h"
#include "ethernet_industry/eth_ecat_command.hpp"
#include "ethernet_industry/eth_ecat_custom_tuple.hpp"

namespace free_rtos {

namespace datagram {

using TEcatWkc = uint16_t;

class IEcatDatagram {
public:
    IEcatDatagram(ec_moredatagrams more, TEcatWkc wkc)
    : more_{more}
    , wkc_{wkc}
    { }

    IEcatDatagram() { }

    IEcatDatagram& operator+(IEcatDatagram &next) {
        more_ = ec_moredatagrams::EC_MOREDATAGRAMS_MORE;
        next_ = &next;

        return next;
    }

    IEcatDatagram* get_next() {
        return next_;
    }

    IEcatDatagram& set_next(IEcatDatagram &next) {

        return operator+(next);
    }

    virtual uint8_t* pack(uint8_t *raw) = 0;
    virtual uint8_t* unpack(uint8_t *raw) = 0;

    TEcatDgHeader& get_header() {
        return header_;
    }

    TEcatWkc& get_wkc() {
        return wkc_;
    }

    // установка размера пробела после поля данных до wkc
    void set_padding(uint16_t padding) {
        padding_ = padding;
    }

protected:
    ec_moredatagrams more_;

    TEcatDgHeader header_;
    TEcatWkc wkc_;
    uint16_t padding_{0x0000};
private:
    IEcatDatagram *next_{nullptr};
};

template<typename CommandT, typename... DataTypes>
class EcatDatagram : public IEcatDatagram {
    static_assert(std::is_base_of<command::EcatCommandBase, CommandT>::value == true, "CommandT should be derived from ECatCommandBase");

public:
    EcatDatagram(CommandT command, DataTypes&... data)
    : IEcatDatagram{ec_moredatagrams::EC_MOREDATAGRAMS_LAST, 0x0000}
    , command_{command}
    , data_{data...} { }

    EcatDatagram() { }

    virtual uint8_t* pack(uint8_t *raw) override {
        return pack_header(raw);
    }

    virtual uint8_t* unpack(uint8_t *raw) override {
        return unpack_header(raw);
    }

private:
    CommandT command_;
    CustomTuple<DataTypes...> data_;

    uint8_t* pack_wkc(uint8_t *raw) {
        TEcatWkc *wkc = new(raw) TEcatWkc{0x0000};

        (void)wkc;

        return raw + sizeof(TEcatWkc);
    }

    uint8_t* pack_padding(uint8_t *raw) {
        std::memset(raw, 0x00, padding_);

        return raw + padding_;
    }

    uint8_t* pack_data(uint8_t *raw) {
        return pack_padding(data_.pack(raw));
    }

    uint8_t* pack_header(uint8_t *raw) {
        uint8_t *data_raw = raw + sizeof(TEcatDgHeader);
        uint8_t *wkc_raw = pack_data(data_raw); // сначала упаковываем  все данные для вычислением их размера
        uint16_t len = wkc_raw - data_raw;      // вычисляем размер данных
        TEcatDgHeader *header_ = new(raw) TEcatDgHeader{
            command_.get_cmd(),
            0x00,
            command_.get_address(),
            {.bits = {
                      .len = len,
                      .c = static_cast<uint16_t>(ec_circframe::EC_CIRCFRAME_NOT),
                      .m = static_cast<uint16_t>(more_)}},
            0x0000};

        (void)header_;

        return pack_wkc(wkc_raw);
    }

    uint8_t* unpack_wkc(uint8_t *raw) {
        TEcatWkc *wkc = reinterpret_cast<TEcatWkc*>(raw);

        wkc_ = *wkc;

        return raw + sizeof(TEcatWkc);
    }

    uint8_t* unpack_padding(uint8_t *raw) {
        return unpack_wkc(raw + padding_);
    }

    uint8_t* unpack_data(uint8_t *raw) {
        return unpack_padding(data_.unpack(raw));
    }

    uint8_t* unpack_header(uint8_t *raw) {
        TEcatDgHeader *header = reinterpret_cast<TEcatDgHeader*>(raw);

        header_ = *header;

        return unpack_data(raw + sizeof(TEcatDgHeader));
    }
};

class EcatTelegram : public Handler {
public:
    EcatTelegram(Eth& eth)
    : eth_{eth}
    , tx_flow_{*eth.getTxFlowPtr()} {
          eth_.getEthStackPtr()->Register(ETH_PROT_ECAT_LE, this);
    }

    virtual int32_t Process(uint8_t *p_data, uint32_t len) override;

    void init(TEthMacPorts port_id) {
        port_id_ = port_id;
    }

    void transfer(IEcatDatagram& first);

private:
    Eth& eth_;
    EthTxFlowIface& tx_flow_;
    TEthMacPorts port_id_;

    free_rtos::Semaphore rx_sem_;

    TEthPkt buffer_out_;
    TEthPkt buffer_in_;

    void pack(IEcatDatagram& first);
    void unpack(IEcatDatagram& first);
};

}


}



#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_ */
